import logging
import re
from dataclasses import dataclass
from functools import cached_property, cmp_to_key
from itertools import groupby
from typing import Iterable

from parse_qwantz.box import Box, get_interval_distance
from parse_qwantz.fonts import Font, CharBox
from parse_qwantz.colors import Color
from parse_qwantz.hyphens import disambiguate_hyphen
from parse_qwantz.text_lines import TextLine, group_text_lines
from parse_qwantz.pixels import Pixel

logger = logging.getLogger()


@dataclass
class TextBlock:
    rows: list[list[TextLine]]
    alignments: list["Alignment"]
    color: Color
    font: Font

    @property
    def start(self) -> Pixel:
        return self.rows[0][0].start

    @property
    def end(self) -> Pixel:
        return self.rows[-1][-1].end

    @property
    def is_bold(self) -> bool:
        return all(line.is_bold for line in self.lines)

    @property
    def is_italic(self) -> bool:
        return all(line.is_italic for line in self.lines)

    @property
    def box(self) -> Box:
        top = self.start.y
        bottom = self.end.y
        left = min(line[0].start.x for line in self.rows)
        right = max(line[-1].end.x for line in self.rows)
        return Box(Pixel(left, top), Pixel(right, bottom))

    @property
    def lines(self) -> Iterable[TextLine]:
        for row in self.rows:
            yield from row

    @cached_property
    def bond_strengths(self) -> list[int]:
        return [alignment.strength for alignment in self.alignments]

    def content(self, mark_bold=True, mark_italic=True, include_font_name=False):
        char_boxes = []
        for row in self.rows:
            row_char_boxes = list(self.get_row_charboxes(row))
            row_content = "".join(char_box.char for char_box in row_char_boxes)
            if char_boxes:
                if char_boxes[-1].char == '-' and char_boxes[-2].char not in ' -':
                    if row_content.endswith("-") and " " not in row_content:
                        logger.warning("Multiple hyphenation")
                    last_words = ''
                    for char_box in char_boxes[-2::-1]:
                        if char_box.char in '.,!?"#(/ ':
                            break
                        last_words = char_box.char + last_words
                    if last_words.startswith("'"):
                        last_words = last_words[1:]
                    next_words = re.match(r'[^].,!?"\' :;)/]*', row_content).group()
                    if not disambiguate_hyphen(last_words.split("-"), next_words.strip("-").split("-")):
                        char_boxes.pop()
                elif (
                    not (row_content.startswith("+") and row_content[1] != " ")
                    and not (len(char_boxes) > 1 and char_boxes[-1].char in "/|" and char_boxes[-2].char != " ")
                    and len(set(cb.char.lower() for cb in char_boxes[-2:]) | set(row_content[:2].lower())) != 1
                    and {char_boxes[-1].char.lower()} | set(row_content[:2].lower()) != {'h'}
                ):
                    char_boxes.append(CharBox.space(is_bold=char_boxes[-1].is_bold, is_italic=char_boxes[-1].is_italic))
            char_boxes.extend(row_char_boxes)

        grouped_char_boxes = groupby(
            char_boxes, key=lambda cb: (cb.is_bold and mark_bold, cb.is_italic and mark_italic)
        )
        text_and_weight = (
            (''.join(char_box.char for char_box in group), is_bold, is_italic)
            for (is_bold, is_italic), group in grouped_char_boxes
        )
        content = ''.join(
            mark_excluding_trailing_spaces(content, '◖◗' if is_bold else '▹◃') if is_bold or is_italic else content
            for content, is_bold, is_italic in text_and_weight
        )
        content = content.replace('  ', ' ')
        if include_font_name and self.font.display_name is not None:
            content = f"〚{self.font.display_name}〛 {content}"
        return content

    @staticmethod
    def get_row_charboxes(row: list[TextLine]) -> Iterable[CharBox]:
        previous_line = None
        for line in row:
            if previous_line and line.box().left - previous_line.box().right >= line.font.space_width // 2:
                yield CharBox.space(is_bold=previous_line.is_bold, is_italic=previous_line.is_italic)
            yield from line.char_boxes
            previous_line = line

    def unambiguous_words(self) -> Iterable[str]:
        row_contents = []
        for row in self.rows:
            row_content = ""
            previous_line = None
            for line in row:
                if previous_line and line.box().left - previous_line.box().right >= line.font.space_width // 2:
                    row_content += " "
                row_content += line.content
                previous_line = line
            row_contents.append(row_content.lower())
        was_hyphen = False
        for row_content in row_contents:
            is_hyphen = row_content[-1] == '-' and row_content[-2] not in ' -'
            words = re.findall(r"\w+", row_content)
            if was_hyphen and row_content[0].isalnum():
                del words[0]
            if is_hyphen and words:
                del words[-1]
            yield from words
            if was_hyphen:
                row_content = "-" + row_content
            yield from re.findall(r"(?<![-\w])\w+(?:-\w+)+(?![-\w])", row_content)
            was_hyphen = is_hyphen

    def split(self, line1: TextLine, line2: TextLine) -> tuple["TextBlock", "TextBlock", "Alignment"]:
        line1_index = self.row_index(line1)
        line2_index = self.row_index(line2)
        index1, index2 = sorted((line1_index, line2_index))
        breaking_strength, split_index = min(
            ((self.bond_strengths[i], i) for i in range(index1, index2)),
            key=lambda p: (p[0], -p[1]),
        )
        if self.bond_strengths[index1:index2].count(breaking_strength) > 1:
            logger.warning(f"Non-unique breaking point for splitting (strength: {breaking_strength})")
        block1 = TextBlock(self.rows[:split_index+1], self.alignments[:split_index], self.color, self.font)
        block2 = TextBlock(self.rows[split_index+1:], self.alignments[split_index+1:], self.color, self.font)
        if line1_index < line2_index:
            return block1, block2, self.alignments[split_index]
        else:
            return block2, block1, self.alignments[split_index]

    def alignment(self, line1: TextLine, line2: TextLine) -> "Alignment":
        line1_index = self.row_index(line1)
        line2_index = self.row_index(line2)
        index1, index2 = sorted((line1_index, line2_index))
        _, split_index = min((self.bond_strengths[i], i) for i in range(index1, index2))
        return self.alignments[split_index]

    def can_split(self, line1: TextLine, line2: TextLine) -> bool:
        if self.row_index(line1) == self.row_index(line2):
            return False
        alignment = self.alignment(line1, line2)
        return not (alignment.no_gap and alignment.left_aligned)

    def row_index(self, line: TextLine) -> int:
        return next(i for i, row in enumerate(self.rows) if line in row)

    def extra_info(self) -> list[str]:
        return [
            char_box.extra_info
            for line in self.lines
            for char_box in line.char_boxes
            if char_box.extra_info
        ]

    def __str__(self):
        return self.content()

    def __hash__(self):
        return id(self)


def compare_text_blocks(block1: TextBlock, block2: TextBlock):
    if block1.end.y - block2.font.height / 2 <= block2.start.y:
        return -1
    if block1.start.y >= block2.end.y - block2.font.height / 2:
        return 1
    if block1.start.x < block2.start.x:
        if block1.start.y - block1.font.height*1.5 <= block2.start.y:
            return -1
        if block1.end.y <= block2.end.y:
            return -1
        logger.warning(f"Ambiguous block order {block1.box} {block2.box} {block1.font.height}")
        return 1
    if block1.start.x > block2.start.x:
        if block1.start.y >= block2.start.y - block2.font.height*1.5:
            return 1
        if block1.end.y >= block2.end.y:
            return 1
        logger.warning(f"Ambiguous block order {block2.box} {block1.box} {block2.font.height}")
        return -1
    return 0


def sort_text_blocks(text_blocks: Iterable[TextBlock]) -> list[TextBlock]:
    return sorted(text_blocks, key=cmp_to_key(compare_text_blocks))


def mark_excluding_trailing_spaces(s: str, markers: str) -> str:
    trailing_spaces = re.search(r" *$", s).group()
    return f'{markers[0]}{s.rstrip()}{markers[1]}{trailing_spaces}'


@dataclass
class Alignment:
    left_aligned: bool
    char_aligned: bool
    no_gap: bool

    @property
    def strength(self):
        bond_strength = 0
        if self.left_aligned:
            bond_strength += 5
        elif self.char_aligned:
            bond_strength += 3
        if self.no_gap:
            bond_strength += 10
        return bond_strength


def get_text_blocks(text_lines: list[TextLine]) -> Iterable[TextBlock]:
    grouped_lines = group_text_lines(text_lines)
    while grouped_lines:
        new_block = [grouped_lines[0]]
        alignments = []
        font = grouped_lines[0][0].font
        new_lines: list[list[TextLine]] = []
        for text_line_group in grouped_lines[1:]:
            alignment = fit_to_block(text_line_group, new_block[-1], font)
            if alignment is not None:
                new_block.append(text_line_group)
                alignments.append(alignment)
            else:
                new_lines.append(text_line_group)
        grouped_lines = new_lines
        color = new_block[0][0].color
        yield TextBlock(new_block, alignments, color, font)


def fit_to_block(line_group: list[TextLine], previous_group: list[TextLine], font: Font) -> Alignment | None:
    if line_group[0].font.group != font.group:
        return None
    if line_group[0].color != previous_group[0].color:
        return None
    first_box = line_group[0].box()
    last_box = line_group[-1].box()
    left = first_box.left
    right = last_box.right
    top = min(first_box.top, last_box.top)
    if all(line.is_bold for line in previous_group) and not any(line.contains_bold for line in line_group):
        return None
    previous_first_box = previous_group[0].box()
    previous_last_box = previous_group[-1].box()
    previous_left = previous_first_box.left
    previous_right = previous_last_box.right
    previous_bottom = max(previous_first_box.bottom, previous_last_box.bottom)
    if get_interval_distance(
        (left, right),
        (previous_left, previous_right),
    ) != 0:
        return None
    previous_height = font.height
    previous_width = font.space_width
    common_part = min(right - previous_left, previous_right - left)
    if common_part < previous_width:
        return None
    if previous_bottom - 1 <= top <= previous_bottom + previous_height // 6:
        return Alignment(
            left_aligned=previous_left == left,
            char_aligned=(previous_left - left) % previous_width == 0 and font.is_mono,
            no_gap=top <= previous_bottom,
        )
    return None



import logging
import math
from importlib.resources import files
from itertools import chain
from typing import Iterator
from xml.etree import ElementTree

from PIL import Image

import parse_qwantz
from parse_qwantz.colors import Color
from parse_qwantz.elements import get_elements, Direction
from parse_qwantz.fonts import CharBox, Font
from parse_qwantz.lines import Line
from parse_qwantz.match_lines import Character
from parse_qwantz.panels import PANELS
from parse_qwantz.pixels import is_ask_professor_science, Pixel
from parse_qwantz.prepare_image import prepare_image
from parse_qwantz.simple_image import SimpleImage

logger = logging.getLogger()

BLANK_FILE_PATH = files(parse_qwantz).joinpath('img/blank.svg')
ASK_PROFESSOR_SCIENCE_PATH = files(parse_qwantz).joinpath('img/ask-professor-science.svg')
BATMAN_LEFT_PATH = files(parse_qwantz).joinpath('img/batman-left.svg')
BATMAN_RIGHT_PATH = files(parse_qwantz).joinpath('img/batman-right.svg')


def generate_svg(image: Image):
    masked, good_panels = prepare_image(image)
    ask_professor_science = _is_ask_professor_science(masked)
    simple_image = SimpleImage.from_image(masked, ask_professor_science)
    lines, char_boxes, characters = get_elements_for_svg(simple_image)
    svg_elements = [make_line_element(line, width, i) for i, (line, width) in enumerate(lines)]
    svg_elements.extend(
        chain(*(make_text_element(char_box, font, color, i) for i, (char_box, font, color) in enumerate(char_boxes)))
    )
    svg_elements.extend((make_character(character) for character in characters))
    if ask_professor_science:
        svg_elements.append(get_professor_science_sign())
    return render_svg(svg_elements)


def _is_ask_professor_science(image: Image) -> bool:
    panel = PANELS[0]
    (width, height), (x, y) = panel
    cropped = image.crop((x, y, x + width, y + height))
    return is_ask_professor_science(cropped)


def get_professor_science_sign() -> ElementTree.Element:
    doc = ElementTree.parse(ASK_PROFESSOR_SCIENCE_PATH)
    root = doc.getroot()
    return root[0]


def get_batman(direction: Direction) -> ElementTree.Element:
    batman_path = BATMAN_LEFT_PATH if direction == Direction.LEFT else BATMAN_RIGHT_PATH
    doc = ElementTree.parse(batman_path)
    root = doc.getroot()
    return root[0]


def render_svg(svg_elements: list[ElementTree.Element]) -> str:
    ElementTree.register_namespace('', "http://www.w3.org/2000/svg")
    ElementTree.register_namespace('rdf', "http://www.w3.org/1999/02/22-rdf-syntax-ns#")
    ElementTree.register_namespace('cc', "http://creativecommons.org/ns#")
    doc = ElementTree.parse(BLANK_FILE_PATH)
    root = doc.getroot()
    root.extend(svg_elements)
    return ElementTree.tostring(root).decode()


def make_line_element(line: Line, width: int, no: int) -> ElementTree.Element:
    line_element = ElementTree.Element(
        "line",
        {
            "id": f"line{no}",
            "x1": str(line[0].x + 0.5),
            "y1": str(line[0].y + 0.5),
            "x2": str(line[1].x + 0.5),
            "y2": str(line[1].y + 0.5),
            "stroke": "black",
            "stroke-width": str(width),
        }
    )
    line_element.tail = "\n"
    return line_element


def make_text_element(char_box: CharBox, font: Font, color: Color, no: int) -> Iterator[ElementTree.Element]:
    font_family = "Times New Roman" if font.group == "TNR13" else "Lucida Console"
    x = char_box.box.left
    if char_box.is_italic:
        x += len(font.italic_offsets) / 2
    if char_box.is_bold:
        xs = [x, x + 0.25, x + 0.5, x + 0.75]
    else:
        xs = [x]
    for x in xs:
        attrib = {
            "id": f"text{no}",
            "x": str(x),
            "y": str(char_box.box.bottom - (font.height - font.base - 1)),
            "fill": f"#{color.red:02X}{color.green:02X}{color.blue:02X}"
        }
        styling = [
            f"font-family: {font_family};",
            f"font-size: {font.height}px;",
        ]
        if char_box.is_italic:
            styling.append("font-style: italic;")
        if styling:
            attrib["style"] = " ".join(styling)
        text_element = ElementTree.Element("text", attrib)
        text_element.text = char_box.char
        text_element.tail = "\n"
        yield text_element


def make_character(character: Character) -> ElementTree.Element:
    assert character.name == "Floating Batman head", character.name
    batman = get_batman(character.direction)
    box = character.boxes[0]
    attrib = {
        "id": "batman",
        "transform": f"translate({box.left},{box.top})"
    }
    group_element = ElementTree.Element("g", attrib)
    group_element.append(batman)
    return group_element


def get_elements_for_svg(
    image: SimpleImage
) -> tuple[list[tuple[Line, int]], list[tuple[CharBox, Font, Color]], list[Character]]:
    lines, line_widths, thoughts, text_lines, extra_characters, unmatched = get_elements(image)
    if thoughts or unmatched:
        logger.warning(f"Foreign elements in the image")
    char_boxes = [
        (char_box, text_line.font, text_line.color)
        for text_line in text_lines
        for char_box in text_line.char_boxes
        if char_box.pixels
    ]
    lines = (fix_for_panel_edges(line) for line in lines)
    return list(zip(lines, line_widths)), char_boxes, extra_characters


def fix_for_panel_edges(line: Line) -> Line:
    start, end = line
    length = math.sqrt((start.x - end.x)**2 + (start.y - end.y)**2)
    return _get_new_end(start, end, length), _get_new_end(end, start, length)


def _get_new_end(end: Pixel, other_end: Pixel, length: float) -> Pixel:
    if any(is_on_edge(end, panel) for panel in PANELS):
        return Pixel(
            x=int(end.x + 2 * (end.x - other_end.x) / length),
            y=int(end.y + 2 * (end.y - other_end.y) / length)
        )
    else:
        return end


def is_on_edge(pixel: Pixel, panel: tuple[tuple[int, int], tuple[int, int]]) -> bool:
    (width, height), (x, y) = panel
    return (
        (x <= pixel.x < x + width and pixel.y in (y, y + height - 1))
        or (y <= pixel.y < y + height and pixel.x in (x, x + width - 1))
    )



import hashlib
import logging
import re
import sys
from pathlib import Path
from typing import Iterable

from PIL import Image, ImageDraw

from parse_qwantz.panel_overrides import get_panel_overrides
from parse_qwantz.panels import PANELS, CHARACTERS
from parse_qwantz.parser import parse_qwantz, match_stuff, parse_footer
from parse_qwantz.elements import get_elements
from parse_qwantz.simple_image import SimpleImage
from parse_qwantz.prepare_image import prepare_image
from parse_qwantz.svg_gen import generate_svg


def get_unambiguous_words(image: Image) -> Iterable[str]:
    md5 = hashlib.md5(image.tobytes()).hexdigest()
    panel_overrides = get_panel_overrides().get(md5, {})
    masked, good_panels = prepare_image(image)
    for i, (panel, characters) in enumerate(zip(PANELS, CHARACTERS), start=1):
        if str(i) in panel_overrides:
            panel = panel_overrides[str(i)]
            for panel_line in panel:
                yield from get_words(panel_line)
            continue
        (width, height), (x, y) = panel
        cropped = masked.crop((x, y, x + width, y + height))
        panel_image = SimpleImage.from_image(cropped)
        lines, _widths, thoughts, text_lines, extra_characters, unmatched_shapes = get_elements(panel_image)
        text_blocks, block_matches, thought_blocks, unmatched_stuff = match_stuff(
            characters + extra_characters, panel_image, lines, text_lines, thoughts
        )
        for text_block in text_blocks:
            yield from text_block.unambiguous_words()


def get_words(line: str) -> list[str]:
    if line[0] == '〚':
        return []
    line_text = line.split(': ', 1)[1].lower()
    line_text = re.sub(r"〚[^〛]*〛", "", line_text)
    line_text = re.sub(r"\w*⦃[^⦄]*⦄\w*", "", line_text)
    line_text = re.sub(r"\w*…\w*", "", line_text)
    return re.findall(r"\w+", line_text) + re.findall(r"\w+(?:-\w+)+", line_text)


def main(
    input_file_path: Path,
    output_dir: Path | None = None,
    debug: bool = False,
    show_boxes: bool = False,
    unambiguous_words: bool = False,
    svg: bool = False,
    footer: bool = False,
):
    image = Image.open(input_file_path)
    if unambiguous_words:
        return list(get_unambiguous_words(image))
    if svg:
        print(generate_svg(image))
        return
    output_file = (output_dir / (input_file_path.name + '.txt')).open('w') if output_dir else sys.stdout
    if output_dir:
        logging.basicConfig(filename=output_dir / (input_file_path.stem + '.log'), filemode='w', force=True)
    if footer:
        for line in parse_footer(image):
            print(line, file=output_file)
        return
    for panel_no, lines in enumerate(parse_qwantz(image, debug=debug, log_colors=not output_dir), start=1):
        for line in lines:
            print(line, file=output_file)
        if panel_no != 6:
            print(file=output_file)
    if show_boxes:
        draw = ImageDraw.Draw(image)
        for (panel, characters) in zip(PANELS, CHARACTERS):
            _, (x, y) = panel
            for character in characters:
                for box in character.boxes:
                    (x0, y0), (x1, y1), _ = box
                    draw.rectangle(((x0 + x, y0 + y), (x1 + x, y1 + y)), outline=(0, 128, 0))
        image.show()
    output_file.flush()



import hashlib
import logging
from dataclasses import dataclass
from typing import Iterable

from PIL import Image, ImageDraw

from parse_qwantz import colors
from parse_qwantz.box import Box, get_interval_distance
from parse_qwantz.color_logs import ColorFormatter
from parse_qwantz.elements import get_elements
from parse_qwantz.lines import Line
from parse_qwantz.match_blocks import match_blocks
from parse_qwantz.match_lines import Character, match_lines, OFF_PANEL
from parse_qwantz.match_thought import match_thought
from parse_qwantz.panels import PANELS, CHARACTERS, FOOTER
from parse_qwantz.panel_overrides import get_panel_overrides
from parse_qwantz.pixels import is_ask_professor_science, Pixel
from parse_qwantz.prepare_image import prepare_image
from parse_qwantz.shape import get_box
from parse_qwantz.simple_image import SimpleImage
from parse_qwantz.text_blocks import TextBlock, sort_text_blocks, get_text_blocks
from parse_qwantz.text_lines import TextLine

logger = logging.getLogger()


def parse_qwantz(
    image: Image, debug: bool = False, log_colors: bool = False, ignore_overrides: bool = False
) -> Iterable[list[str]]:
    md5 = hashlib.md5(image.tobytes()).hexdigest()
    panel_overrides = get_panel_overrides().get(md5, {}) if not ignore_overrides else {}
    masked, good_panels = prepare_image(image)
    for i, (panel, characters) in enumerate(zip(PANELS, CHARACTERS), start=1):
        if str(i) in panel_overrides:
            yield panel_overrides[str(i)]
            continue
        if i not in good_panels:
            logger.warning("Non-standard panel without an override")
            continue
        set_current_panel(i, log_colors)
        (width, height), (x, y) = panel
        cropped = masked.crop((x, y, x + width, y + height))
        ask_professor_science = is_ask_professor_science(cropped)
        panel_image = SimpleImage.from_image(cropped, ask_professor_science)
        lines, _widths, thoughts, text_lines, extra_characters, unmatched_shapes = get_elements(panel_image)
        text_blocks, block_matches, thought_blocks, unmatched_stuff = match_stuff(
            characters + extra_characters, panel_image, lines, text_lines, thoughts
        )
        for text_block in text_blocks:
            for extra_info in text_block.extra_info():
                logger.warning(f"Variant used: {extra_info}")
        script_lines = list(get_script_lines(text_blocks, block_matches, thought_blocks, ask_professor_science))
        if debug and (unmatched_shapes or unmatched_stuff):
            handle_debug(cropped, text_blocks, unmatched_shapes, unmatched_stuff, characters + extra_characters)
        if script_lines:
            yield list(script_lines)
        else:
            yield ["〚no text〛"]


def parse_footer(image: Image) -> list[str]:
    md5 = hashlib.md5(image.tobytes()).hexdigest()
    panel_overrides = get_panel_overrides().get(md5, {})
    if "footer" in panel_overrides:
        return panel_overrides["footer"]
    masked, _ = prepare_image(image)
    (width, height), (x, y) = FOOTER
    cropped = masked.crop((x, y, x + width, y + height))
    footer_image = SimpleImage.from_image(cropped)
    lines, _widths, thoughts, text_lines, extra_characters, unmatched_shapes = get_elements(footer_image)
    if lines or thoughts or extra_characters or unmatched_shapes:
        logger.warning("Unexpected elements in footer")
    for text_line in text_lines:
        for char_box in text_line.char_boxes:
            if char_box.extra_info:
                logger.warning(f"Variant used: {char_box.extra_info}")
    return [text_line.content for text_line in text_lines]


def set_current_panel(panel: int, use_colors: bool = True):
    panel_name = f" Panel {panel}:"
    if not logger.handlers:
        logging.basicConfig()
    logger.handlers[0].setFormatter(ColorFormatter(defaults={"panel": panel_name}, colors=use_colors))


@dataclass
class UnmatchedStuff:
    neighbors: list[tuple[TextLine, TextLine]]
    lines: list[Line]
    thoughts: list[Box]

    def __bool__(self):
        return bool(self.neighbors or self.lines or self.thoughts)


def match_stuff(
    characters: list[Character], image: SimpleImage, lines: list[Line], text_lines: list[TextLine], thoughts: list[Box]
) -> tuple[
    list[TextBlock],
    dict[TextBlock, list[Character]],
    list[TextBlock],
    UnmatchedStuff,
]:
    text_blocks = sort_text_blocks(get_text_blocks(text_lines))
    line_matches, unmatched_lines = match_lines(lines, text_blocks, characters, image)
    block_matches, text_blocks, unmatched_neighbors = match_blocks(line_matches, text_blocks)
    text_blocks = sort_text_blocks(text_blocks)
    unmatched_blocks = [block for block in text_blocks if block not in block_matches]
    thought_matches = list(match_thought(thoughts, unmatched_blocks))
    if thoughts and not thought_matches:
        logger.warning("Detected thought bubbles, but no thought text")
        unmatched_thoughts = thoughts
    else:
        unmatched_thoughts = []
    unmatched_blocks = [block for block in unmatched_blocks if block not in thought_matches]
    match_above_or_below(unmatched_blocks, block_matches)
    unmatched_stuff = UnmatchedStuff(unmatched_neighbors, unmatched_lines, unmatched_thoughts)
    return text_blocks, block_matches, thought_matches, unmatched_stuff


def get_script_lines(
    text_blocks: list[TextBlock],
    block_matches: dict[TextBlock, list[Character]],
    thought_blocks: list[TextBlock],
    ask_professor_science: bool,
) -> Iterable[str]:
    if ask_professor_science:
        yield "Sign: ASK PROFESSOR SCIENCE"
    for block in text_blocks:
        if god_or_devil := handle_god_and_devil(block, block_matches.get(block) == [OFF_PANEL]):
            block_matches[block] = [god_or_devil]
        if block in block_matches:
            characters = block_matches[block]
            if god_or_devil:
                if god_or_devil.name in ("God", "Devil"):
                    content = block.content(mark_bold=False)
                else:
                    content = block.content(mark_italic=False)
            elif characters[0].name == "Floating Batman head":
                content = block.content()
            else:
                content = block.content(include_font_name=True)
            line = f"{' and '.join(ch.name for ch in characters)}: {content}"
            if not line[0].isupper():
                line = line[0].upper() + line[1:]
            yield line
        elif block in thought_blocks:
            yield f"T-Rex: 〚thinks〛 {block.content()}"
        elif not block.font.is_mono:
            yield f"Text: {block.content()}"
        else:
            if not block.is_bold:
                logger.warning('Narrator not bold: %s', block.font.name)
            yield f"Narrator: {block.content(mark_bold=False)}"


def handle_debug(
    image: Image,
    text_blocks: list[TextBlock],
    unmatched_shapes: list[list[Pixel]],
    unmatched_stuff: UnmatchedStuff,
    characters: list[Character],
):
    draw = ImageDraw.Draw(image)
    for unmatched_shape in unmatched_shapes:
        box = get_box(unmatched_shape, padding=3)
        draw.rectangle(box[:2], outline=(255, 0, 0))
        for pixel in unmatched_shape:
            draw.point(pixel, fill=(255, 0, 0))
    for thought_box in unmatched_stuff.thoughts:
        draw.rectangle(thought_box[:2], outline=(255, 0, 255))
    for text_line1, text_line2 in unmatched_stuff.neighbors:
        box1 = text_line1.box()
        box2 = text_line2.box()
        draw.rectangle(box1[:2], outline=(0, 0, 192))
        draw.rectangle(box2[:2], outline=(0, 0, 192))
        draw.line([
            ((box1.left + box1.right) // 2, (box1.top + box1.bottom) // 2),
            ((box2.left + box2.right) // 2, (box2.top + box2.bottom) // 2),
        ], fill=(0, 0, 255))
    for line in unmatched_stuff.lines:
        draw.line(line, fill=(255, 0, 0))
        for block in text_blocks:
            for text_line in block.lines:
                box = text_line.box()
                draw.rectangle(box[:2], outline=(0, 192, 0))
    if unmatched_stuff.neighbors or unmatched_stuff.lines:
        for character in characters:
            for box in character.boxes:
                draw.rectangle(box[:2], outline=(0, 128, 0))
    image.show()


def match_above_or_below(unmatched_blocks: list[TextBlock], block_matches: dict[TextBlock, list[Character]]) -> None:
    for unmatched_block in unmatched_blocks:
        box = unmatched_block.box
        closest = None
        best_distance = None
        if not unmatched_block.is_bold:
            for block, characters in block_matches.items():
                if not characters:
                    continue
                if block.font.group != unmatched_block.font.group or block.color != unmatched_block.color:
                    continue
                other_box = block.box
                if get_interval_distance((box.left, box.right), (other_box.left, other_box.right)) == 0:
                    distance = max(other_box.top - box.bottom, box.top - other_box.bottom)
                    line_height = max(unmatched_block.font.height, block.font.height)
                    if distance < line_height * 2 and (best_distance is None or distance < best_distance):
                        closest = characters
                        best_distance = distance
        if closest:
            logger.warning("Matching disconnected blocks")
            block_matches[unmatched_block] = closest


def handle_god_and_devil(block: TextBlock, is_off_panel: bool) -> Character | None:
    if any(char.islower() for char in block.content()):
        return None
    if block.color == colors.RED and is_off_panel and block.is_bold:
        return Character.from_name('Devil')
    elif is_off_panel and block.is_bold:
        return Character.from_name('God')
    elif is_off_panel and block.is_italic:
        return Character.from_name('Creepy voice(s)')



import logging
from dataclasses import dataclass
from multiprocessing import Pool
from pathlib import Path

import typer
from PIL import ImageShow

from parse_qwantz.color_logs import set_logging_formatter
from parse_qwantz.image_viewer import SilentViewer
from parse_qwantz.main import main
from parse_qwantz.prepare_image import ImageError

ImageShow.register(SilentViewer(), 0)

app = typer.Typer()


@dataclass
class Inner:
    output_dir: Path | None
    debug: bool
    show_boxes: bool
    unambiguous_words: bool
    generate_svg: bool
    parse_footer: bool

    def __call__(self, image_path: Path):
        try:
            return main(
                image_path,
                output_dir=self.output_dir,
                debug=self.debug,
                show_boxes=self.show_boxes,
                unambiguous_words=self.unambiguous_words,
                svg=self.generate_svg,
                footer=self.parse_footer
            )
        except ImageError:
            pass


@app.command()
def cli(
    input_paths: list[Path] = typer.Argument(..., help="Paths to one or more image and/or directory", exists=True),
    output_dir: Path = typer.Option(None, help="Path to the output directory", exists=True, file_okay=False),
    log_level: str = typer.Option('WARNING', help="Log level"),
    debug: bool = typer.Option(False, help="Enable debug features."),
    show_boxes: bool = typer.Option(False, help="Show character boxes (for debug)"),
    unambiguous_words: bool = typer.Option(False, help="Print only unambiguous words"),
    generate_svg: bool = typer.Option(False, help="Generate SVG file"),
    parse_footer: bool = typer.Option(False, help="Parse the footer rather then the comic"),
):
    """Generate transcripts for Ryan North's Dinosaur Comics from https://qwantz.com"""
    set_logging_formatter()
    logger = logging.getLogger()
    logger.setLevel(getattr(logging, log_level.upper()))
    for input_path in input_paths:
        if input_path.is_dir():
            image_paths = (child_path for child_path in input_path.iterdir() if child_path.is_file())
        else:
            image_paths = [input_path]

        inner = Inner(
            output_dir=output_dir,
            debug=debug,
            show_boxes=show_boxes,
            unambiguous_words=unambiguous_words,
            generate_svg=generate_svg,
            parse_footer=parse_footer,
        )

        with Pool(4) as pool:
            results = pool.map(inner, image_paths)
            if unambiguous_words:
                for result in results:
                    for word in result:
                        print(word)


if __name__ == '__main__':
    app()
